require "class"

local gui = require "quickie"
local Demo = require "demo"

function love.load()
    local glsl = {
        hsv = love.filesystem.read("effects/hsv.glsl"),

        random2d  = love.filesystem.read("effects/random2d.glsl"),  --  
        perlin3d  = love.filesystem.read("effects/perlin3d.glsl"),  -- -1 to 1 ?
        simplex3d = love.filesystem.read("effects/simplex3d.glsl"), -- -1 to 1 ?
        worley3d  = love.filesystem.read("effects/worley3d.glsl"),  --  0 to 1 ?
        flow2d    = love.filesystem.read("effects/flow2d.glsl"),    -- -1 to 1 ?
    }

    Perlin = Demo(
        love.graphics.newPixelEffect(glsl.perlin3d .. glsl.hsv .. [[
            extern number width, height;

            extern number zoom, time;
            extern number ox, oy;

            extern number hue, saturation;

            vec4 effect(vec4 colour, Image image, vec2 local, vec2 screen)
            {
                vec2 camera = vec2(ox, oy);
                vec2 point  = screen - vec2(width, height) * 0.5 + camera;

                point.x /= width;
                point.y /= height;

                vec3 coords = vec3(point / zoom + camera, time);

                number noise = perlin3d(coords) * 0.5 + 0.5;

                return vec4(hsv(hue, saturation, noise), 1.0);
            }
        ]]),

        "Bare 3D Perlin noise (x, y, time)"
    )

    Simplex = Demo(
        love.graphics.newPixelEffect(glsl.simplex3d .. glsl.hsv .. [[
            extern number width, height;

            extern number zoom, time;
            extern number ox, oy;

            vec4 effect(vec4 colour, Image image, vec2 local, vec2 screen)
            {
                vec2 camera = vec2(ox, oy);
                vec2 point  = screen - vec2(width, height) * 0.5 + camera;

                point.x /= width;
                point.y /= height;

                vec3 coords = vec3(point / zoom + camera, time);

                number noise = simplex3d(coords) * 0.5 + 0.5;

                return vec4(vec3(noise), 1.0);
            }
        ]]),

        "Bare 3D Simplex noise (x, y, time)"
    )

    Worley = Demo(
        love.graphics.newPixelEffect(glsl.worley3d .. glsl.hsv .. [[
            extern number width, height;

            extern number zoom, time;
            extern number ox, oy;

            extern number hue, saturation;

            vec4 effect(vec4 colour, Image image, vec2 local, vec2 screen)
            {
                vec2 camera = vec2(ox, oy);
                vec2 point  = screen - vec2(width, height) * 0.5 + camera;

                point.x /= width;
                point.y /= height;

                vec3 coords = vec3(point / zoom + camera, time);

                number noise = worley3d(coords).x;

                return vec4(hsv(hue, saturation, noise), 1.0);
            }
        ]]),

        "Bare 3D Worley noise (x, y, time)"
    )

    Flow = Demo(
        love.graphics.newPixelEffect(glsl.flow2d .. glsl.hsv .. [[
            extern number width, height;

            extern number zoom, time;
            extern number ox, oy;

            vec4 effect(vec4 colour, Image image, vec2 local, vec2 screen)
            {
                vec2 camera = vec2(ox, oy);
                vec2 point  = screen - vec2(width, height) * 0.5 + camera;

                point.x /= width;
                point.y /= height;

                vec2 coords = point / zoom + camera;
                vec2 gradient;

                number noise = flow2d(coords, time, gradient) * 0.5 + 0.5;

                return vec4(vec3(noise), 1.0);
            }
        ]]),

        "Bare 2D Flow noise ((x, y), time)"
    )

    Random = Demo(
        love.graphics.newPixelEffect(glsl.random2d .. glsl.hsv .. [[
            extern number width, height;

            extern number zoom, time;
            extern number ox, oy;

            extern number quantise;

            vec2 quantize(vec2 coords, number y) {
                return floor(coords * y) / y;
            }

            vec4 effect(vec4 colour, Image image, vec2 local, vec2 screen)
            {
                vec2 camera = vec2(ox, oy);
                vec2 point  = screen - vec2(width, height) * 0.5 + camera;

                point.x /= width;
                point.y /= height;

                vec2 coords = vec2(point / zoom + camera);
                coords = quantize(coords, pow(quantise, 2.0) * 32.0);
                coords += vec2(time, time);

                number noise = random2d(coords) * 0.5 + 0.5;

                return vec4(vec3(noise), 1.0);
            }
        ]]),

        "Random noise",

        {name = "quantise", value = 1.0, type = "slider"}
    )

    Rainbow = Demo(
        love.graphics.newPixelEffect(glsl.perlin3d .. glsl.hsv .. [[
            extern number width, height;

            extern number zoom, time;
            extern number ox, oy;
            extern number vx, vy;

            vec4 effect(vec4 colour, Image image, vec2 local, vec2 screen)
            {
                vec2 camera = vec2(ox, oy);
                vec2 point  = screen - vec2(width, height) * 0.5 + camera;

                point.x /= width;
                point.y /= height;

                vec3 coords = vec3((point / zoom + camera) * vec2(vx * 0.25, 2.0 * vy), time);

                coords.x += perlin3d(vec3(coords.y, 0.0, time * 0.25));

                number hue = mod(coords.x, 1.0);
                
                vec3 rain = hsv(hue, 0.35, 1.0);

                return vec4(rain, 1.0);
            }
        ]]),

        "Oil Slick Rainbow - Uniform hue shifted by 2D Perlin noise (y, time)",

        {name = "vx", value = 0.50, type = "slider"},
        {name = "vy", value = 0.25, type = "slider"}
    )

    Water = Demo(
        love.graphics.newPixelEffect(glsl.worley3d .. glsl.hsv .. [[
            extern number width, height;

            extern number zoom, time;
            extern number ox, oy;

            extern number exponent, hue;

            vec4 effect(vec4 colour, Image image, vec2 local, vec2 screen)
            {
                vec2 camera = vec2(ox, oy);
                vec2 point  = screen - vec2(width, height) * 0.5 + camera;

                point.x /= width;
                point.y /= height;

                vec3 coords = vec3(point / zoom + camera, time);

                number noise = 1.0 - pow(worley3d(coords).x, exponent * 10.0);

                return vec4(hsv(hue, 0.30 + 0.70 * noise, 1.0 - noise * 0.2), 1.0);
            }
        ]]),

        "Tropical Water - Saturation and value depend on 3D Worley noise (x, y, time)",

        {name = "exponent", value = 0.2,  type = "slider"},
        {name = "hue",      value = 0.52, type = "slider"}
    )

    FractalPerlin = Demo(
        love.graphics.newPixelEffect(glsl.perlin3d .. glsl.hsv .. [[
            extern number width, height;

            extern number zoom, time;
            extern number ox, oy;

            extern number layers;
            extern number quantise;
            extern number hue;

            number quantize(number x, number y) {
                return floor(x * y) / y;
            }

            vec4 effect(vec4 colour, Image image, vec2 local, vec2 screen)
            {
                vec2 camera = vec2(ox, oy);
                vec2 point  = screen - vec2(width, height) * 0.5 + camera;

                point.x /= width;
                point.y /= height;

                vec3 coords = vec3(point / zoom + camera, time * 0.05);

                coords.xy *= 0.1;

                number noise = 0;

                number layers = floor(layers * 6.0 + 1.0);

                for (int layer = 1; layer <= layers; ++layer) {
                    noise += 1.0 / pow(2.0, float(layer)) * perlin3d(coords * pow(2.0, float(layer)));
                }

                noise = quantize(noise, pow(quantise, 2.0) * 255.0);

                if (hue > 0.0) {
                    return vec4(hsv(mod(noise, 1.0), 0.5, 1.0), 1.0);
                } else {
                    return vec4(vec3(noise * 0.5 + 0.5), 1.0);
                }
            }
        ]]),

        "Fractal Perlin noise (x, y, time)",

        {name = "layers", value = 0.0, type = "slider"},
        {name = "quantise", value = 1.0, type = "slider"},
        {name = "hue", value = false, type = "checkbox"}
    )

    RandomRainbow = Demo(
        love.graphics.newPixelEffect(glsl.random2d .. glsl.perlin3d ..  glsl.hsv .. [[
            extern number width, height;

            extern number zoom, time;
            extern number ox, oy;

            extern number quantise;
            extern number noise_factor;

            vec2 quantize(vec2 coords, number y) {
                return floor(coords * y) / y;
            }

            vec4 effect(vec4 colour, Image image, vec2 local, vec2 screen)
            {
                vec2 camera = vec2(ox, oy);
                vec2 point  = screen - vec2(width, height) * 0.5 + camera;

                point.x /= width;
                point.y /= height;

                vec2 coords2 = vec2(point / zoom + camera);
                coords2 = quantize(coords2, pow(quantise, 2.0) * 32.0);
                coords2 += quantize(vec2(time, 0) * 0.01, 5000.0);
                number noise = random2d(coords2) * 0.5 + 0.5;

                vec3 coords = vec3(coords2 * vec2(0.50 * 0.25, 2.0 * 0.25), time);
                coords.x += perlin3d(vec3(coords.y, 0.0, time * 0.25));
                number hue = mod(coords.x, 1.0);
                vec3 rain = hsv(hue, 0.35, 1.0 - noise * noise_factor);

                return vec4(rain, 1.0);
            }
        ]]),

        "Rainbow pixelated noise",

        {name = "quantise", value = 1.0, type = "slider"},
        {name = "noise_factor", value = 1.0, type = "slider"}
    )

    effect = 1
    effects = {Perlin, Simplex, Worley, Flow, Random, FractalPerlin, Rainbow, Water, RandomRainbow}

    time = 0
    zoom = 0.1
    ox, oy = 0, 0
    px, py = 0, 0

    shapes = false
end

function love.update(dt)
    time = time + dt

    if love.mouse.isDown("r") then
        local mx, my = love.mouse.getPosition()
        local dx, dy = px - mx, py - my

        ox = ox + dx / zoom / love.graphics.getWidth()
        oy = oy - dy / zoom / love.graphics.getHeight()
    end

    px, py = love.mouse.getPosition()

    local effect = effects[effect]

    effect.effect:send("zoom", zoom)
    effect.effect:send("time", time)
    effect.effect:send("ox", ox)
    effect.effect:send("oy", oy)

    effect:update(dt)
end

function love.draw()
    if stencil then
        love.graphics.setStencil(function()
            love.graphics.rectangle("fill", 50, 100, 150, 150)
            love.graphics.circle("fill", 400, 250, 90, 32)
            love.graphics.triangle("fill", 250, 412, 300, 300, 150, 300)
        end)
    else
        love.graphics.setStencil()
    end

    love.graphics.setPixelEffect(effects[effect].effect)
    love.graphics.setColor(255, 255, 255, 255)
    love.graphics.rectangle("fill", 0, 0, love.graphics.getWidth(), love.graphics.getHeight())

    love.graphics.setStencil()

    love.graphics.setPixelEffect()
    love.graphics.setColor(255, 255, 255, 255)
    gui.core.draw()
end

function love.mousepressed(x, y, button)
    if button == "wu" then
        zoom = zoom * (1 / 0.9)
    elseif button == "wd" then
        zoom = zoom * (0.9 / 1)
    end
end

function love.keypressed(key)
    if key == " " then
        effect = effect + 1

        if effect > #effects then effect = 1 end
    elseif key == "return" then
        stencil = not stencil
    end
end
